
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity jtag_programmer_avalon is
	Generic (
		JTAG_IR_size 		: integer := 5;
		Data_Length_Bits	: integer := 8;
		Clock_Divider_Bits 	: integer :=2 -- 2,3,4,5...
		);
	Port ( 
		--Avalon signals
		avalon_clk 			: in std_logic;						
		avalon_reset 		: in std_logic;						
		avalon_chipselect 	: in std_logic;						
		avalon_address 		: in std_logic_vector(1 downto 0); 	
		avalon_write 		: in std_logic;						
		avalon_writedata 	: in std_logic_vector(31 downto 0); 
		avalon_read 		:  in std_logic;					
		avalon_readdata 	: out std_logic_vector(31 downto 0);
		
		-- JTAG outputs
		TMS_out : out std_logic;
		TDI_out : out std_logic;
		TCK_out : out std_logic;
		TDO_in 	: in std_logic;
		RSTB_out: out std_logic
		);
end jtag_programmer_avalon;

architecture Behavioral of jtag_programmer_avalon is
	-- state machines' signals
	type Program_type is (ST0, ST1, ST2, ST3, ST4, ST5, ST6, ST7, ST8, ST9, ST10, ST100, ST11, ST12, P0); 	 		-- states for the JTAG programming sequencer: 0-12 for programming; P0 - pause for receiving a new word
	signal Program: Program_type;
	type Program_shift is (Sin, Sout, Sset1, Sset2);
	signal Shift: Program_shift;
	
	--signal stop
	signal counter: std_logic_vector((Data_Length_Bits-1) downto 0):= (others => '0');   		-- for controlling number of bits sent to each  JTAG register
	constant reg_size: std_logic_vector(Data_Length_Bits-1 downto 0):= std_logic_vector(to_unsigned((JTAG_IR_size-1),Data_Length_Bits));
	signal counter_divider: std_logic_vector((Clock_Divider_Bits-1) downto 0):= (others => '0'); 			-- counter for reducing the frequency of the main clock
	
	signal busy: std_logic:= '0';																			-- active high when programmin procedure is running
	signal intTCK: std_logic:= '0';
	
	signal clock_enable_1stEdge : std_logic:= '0';															-- signals gating the main clock 
	signal clock_enable_2ndEdge, clock_enable_pause : std_logic:= '1';
	
	-- state flags
	signal	f_idle				: std_logic;	-- for reading
	signal	f_wait_next_word	: std_logic;	
	signal	f_reset_read		: std_logic;	
	
	signal	f_do				: std_logic;	-- for writing
	signal	f_next_word_ready	: std_logic:='0';	
	signal	f_reset_write		: std_logic;	
	
	-- internal registers
	signal data_length	: std_logic_vector(Data_Length_Bits-1 downto 0);
	signal reg_sel 		: std_logic_vector(JTAG_IR_size-1 downto 0);
	signal barrel		: std_logic_vector(31 downto 0):= (others => '0');
	
	signal f_wait_next_word_ext :std_logic;
	signal f_next_word_ready_ext :std_logic;
	signal f_mem:std_logic;
begin
	
	TCK_out <= intTCK;
	RSTB_out <= not f_reset_write;
	f_idle <= not busy;
	
	Program_sequence: process (avalon_clk, avalon_reset, f_reset_write, clock_enable_1stEdge, avalon_read, avalon_write)
	begin
		if avalon_reset = '1' then
			Program <= ST0;
			Shift <= Sset1;
			busy <='0';
			f_wait_next_word <= '0';
			barrel <= (others => '0');
			data_length <= (others => '0');
			reg_sel <= (others => '0');
			TMS_out <= '0';
			counter <= (others => '0');
			
		else
			
			if avalon_clk'event and avalon_clk = '1' then
				
				
				f_mem <= f_wait_next_word;
				
				if  f_wait_next_word = '1' and f_mem = '0'then 		-- if rising edge on 'waiting'
					f_wait_next_word_ext <= '1';			   		-- set external 'waiting'
				elsif f_wait_next_word = '0' and f_mem = '1'then    -- if falling edge on 'waiting' = accepting the new word
					f_next_word_ready <= '0';						-- reset internal 'ready'
					f_next_word_ready_ext <= '0';					-- and REMOVE external 'ready'
				elsif f_next_word_ready_ext = '1' then				-- if external ready
					f_wait_next_word_ext <= '0';					-- cancel exteranal 'witing'
					f_next_word_ready <= '1';						-- and set internal ready
				end if;
						
			--				
			-- GENERATE TCK
			if (clock_enable_2ndEdge = '1' or clock_enable_1stEdge = '1') and busy = '1' and (clock_enable_pause= '1')  then
				if clock_enable_1stEdge = '1' then 
					intTCK <= '0';
				elsif clock_enable_2ndEdge = '1' then
					intTCK <= '1';
				else
					intTCK <= intTCK;
				end if;
			end if;	
			
			-- AVALON interface
			if avalon_chipselect = '1' and avalon_write = '1' then		-- when valid write
				case avalon_address is									-- decode address
					when "00" =>
					barrel(31 downto 0) <= avalon_writedata;
					when "01" =>
					reg_sel <= avalon_writedata(JTAG_IR_size-1 downto 0);
					when "10" =>
					data_length <= avalon_writedata(Data_Length_Bits-1 downto 0);
					when "11" =>
					f_do 				<= avalon_writedata(0);
					f_next_word_ready_ext <= avalon_writedata(1);
					f_reset_write 		<= avalon_writedata(2);
					when others =>
					null;
				end case;
			elsif avalon_chipselect = '1' and avalon_read = '1' then	  -- when valid read
				case avalon_address is									  -- decode address
					when "00" =>
					avalon_readdata <= (barrel(31 downto 0));
					when "11" =>
					avalon_readdata(0) <= f_idle;
					avalon_readdata(1) <= f_wait_next_word_ext;
					avalon_readdata(2) <= f_reset_read;
					avalon_readdata(31 downto 3) <= (others => '0');
					when others =>
					avalon_readdata <= (others => '0');
				end case;
			else
				null;
			end if;
			-- AVALON interface END
			
			-- READ data in and out
			if (clock_enable_1stEdge = '1' or clock_enable_2ndEdge = '1') and (Program = ST11 or Program = ST12) then
				case Shift is
					when Sset1 =>
					Shift <= Sout;
					when Sin =>
					Shift <= Sout;
					if clock_enable_pause = '1' then
						barrel(conv_integer(counter(4 downto 0)-1)) <= TDO_in;
					end if;
					when Sout =>
					Shift <= Sin;
					if clock_enable_pause = '1' then
						TDI_out <= barrel(conv_integer(counter(4 downto 0)));
					else
						TDI_out <= barrel(conv_integer(counter(4 downto 0)-1));
					end if;
					when others =>
					null;
				end case;
			else
				null;
			end if;
			
			-- JTAG state machine
			if clock_enable_1stEdge = '1' then 
				if 	f_reset_write = '1' then
					f_wait_next_word <= '0';
					busy <= '0';
					Program <= ST0;
					Shift <= Sout;
					counter <= (others => '0');
					f_reset_read <= '1';
					TMS_out <= '1';
				else	
					f_reset_read <= '0';
					case Program is
						when ST0 =>
						if f_do = '0'  then							-- stay in the idle state
							Program <= ST0;				 
							busy <= '0';
							TMS_out <= '1';
							f_wait_next_word <= '0';
						elsif f_do = '1' then					
							busy <= '1';
							TMS_out <= '0';
							Program <= ST1;
							Shift <= Sset1;
							clock_enable_pause <= '1';
							f_wait_next_word <= '0';
						else
							null;
						end if;
						
						when ST1 =>
						TMS_out <= '1';
						TDI_out <= '1';
						Program <= ST2;
						
						when ST2 =>
						TMS_out <= '1';
						Program <= ST3;
						
						when ST3 =>
						TMS_out <= '0';
						Program <= ST4;
						
						when ST4 =>
						TMS_out <= '0';
						Program <= ST5;
						counter <= (others => '0');
						
						when ST5 =>								-- choose jtag IR 
						counter <= counter +1;
						if counter < data_length then --stop then
							TMS_out <= '0';
							TDI_out <= reg_sel(conv_integer(counter) mod JTAG_IR_size);
							Program <= ST5;
						else
							TMS_out <= '1';
							TDI_out <= reg_sel(conv_integer(counter) mod JTAG_IR_size);
							Program <= ST6;
						end if;
						
						when ST6 =>
						TMS_out <= '1';
						Program <= ST7;
						
						when ST7 =>
						TMS_out <= '0';
						Program <= ST8;
						
						when ST8 =>
						TMS_out <= '1';
						Program <= ST9;
						
						when ST9 =>
						TMS_out <= '0';
						Program <= ST10;
												
						
						when ST10 =>
						TMS_out <= '0';
						Program <= ST100;
						
						when ST100 =>
						TMS_out <= '0';
						Program <= P0;
						f_wait_next_word <='1';
						clock_enable_pause <= '0';
						counter <= std_logic_vector(to_unsigned(1,Data_Length_Bits));
						
						when ST11 =>					  		-- send data to jtag DR
						if clock_enable_pause = '0' then
							clock_enable_pause <= '1';
						else
							counter <= counter +1;
							if counter < data_length then --stop then
								TMS_out <= '0';
								if counter(4 downto 0)="00000" and counter > std_logic_vector(to_unsigned(1,Data_Length_Bits))  then	  -- pause and wait for the next 32-bit word
									Program <= P0;
									f_wait_next_word <='1';
									clock_enable_pause <= '0';
								else
									Program <= ST11;
									clock_enable_pause <= '1';										
								end if;	
							else
								TMS_out <= '1';				 -- send the last bit together with TMS='1'
								counter <= counter +1;
								Program <= ST12;
							end if;
						end if;
						when P0 =>							-- wait fore the next word
						if f_next_word_ready = '1' then
							Program <= ST11;
							f_wait_next_word <= '0';
						else
							Program <= P0;
							f_wait_next_word <= '1';
						end if;
						
						when ST12 =>
						TMS_out <= '1';
						Program <= ST0;
						
						when others =>
						null;
					end case;
				end if;	
			end if;
			else			
			null;		
		end if;	
		
		end if;
	end process;
	
	
	-- "CLOCK divider" - process that generates clock_enables for the JTAG state machine
	generate_enable: process(avalon_clk, avalon_reset)
	begin
		if avalon_reset = '1'then
			counter_divider <= (others => '0');
			clock_enable_1stEdge <= '0';
			clock_enable_2ndEdge <= '0';
		else
			if avalon_clk'event and avalon_clk = '0' then
				counter_divider <= counter_divider + 1;
				if counter_divider = std_logic_vector(to_unsigned(0,Clock_Divider_Bits)) then 			   						-- if 0
					clock_enable_1stEdge <= '1';
					clock_enable_2ndEdge <= '0';
				elsif counter_divider = std_logic_vector(to_unsigned(2**(Clock_Divider_Bits-1),Clock_Divider_Bits)) then		-- if "100..0"
					clock_enable_1stEdge <= '0';
					clock_enable_2ndEdge <= '1';
				else
					clock_enable_1stEdge <='0';
					clock_enable_2ndEdge <= '0';
				end if;
			end if;
		end if;
	end process;
end Behavioral;